#include <inttypes.h>

enum {
    STORAGE_static = 1,
    STORAGE_extern = 2,
    STORAGE_inline = 4,
    STORAGE_typedef = 8,
    STORAGE_mask = 15,
    QUALIFY_const = 16,
    QUALIFY_volatile = 32,
    QUALIFY_restrict = 64,
    TYPE_pointer = 1 << 8,
    TYPE_array = 2 << 8,
    TYPE_function = 3 << 8,
    TYPE_struct = 4 << 8,
    TYPE_union = 5 << 8,
    TYPE_enum = 6 << 8,
    TYPE_ellipsis = 7 << 8,
    TYPE_typedef = 8 << 8,
};

struct Type;

typedef struct Arg {
    struct Type* type;
    struct Arg* next;
} Arg;

typedef struct {
    struct Type* ret;
    Arg* args;
} Func;

typedef struct Member {
    uint32_t tok;
    struct Type* type;
    char* name;
    int bit;
    struct Member* next;
} Member;

/* it is used for union, too */
typedef struct {
    Member* mems;
    char* name;
} Struct;

enum { HASH_SIZE = 8192 };
typedef struct Sym {
    uint32_t tok;
    struct Type* sym;
    char* name;
    Struct* str;
} Sym;

typedef struct Type {
    void* type;
    int qual;
} Type;

enum {
    PRIM_schar = 129,
    PRIM_uchar,
    PRIM_ushort,
    PRIM_uint,
    PRIM_ulong,
    PRIM_longlong,
    PRIM_ulonglong,
    PRIM_longdouble,
    PRIM_int,
    PRIM_void,
    PRIM_char,
    PRIM_short,
    PRIM_long,
    PRIM_float,
    PRIM_double,
};

void chop_init();

void chop_add_include(const char* p);

int chop_cparse(const char* filename);

Sym* chop_get_symbol(const char* name);
char* chop_get_macro(const char* name);

int chop_is_primitive(const Type* type);

void chop_iter_symbols(void (*callback) (const char* name, const Type* type));
void chop_iter_macros(void (*callback) (const char* name, const char* value));

/* should free return value */
char* chop_type2str(const Type* type);

void chop_quit();
