#include <ruby.h>

#include "../cparser.h"

static VALUE rb_include(VALUE k, VALUE s) {
    int ret;
    StringValue(s);
    ret = chop_cparse(RSTRING(s)->ptr);
    if (ret) return Qnil;
    else return Qtrue;
}

static VALUE type2str(Type* type) {
    type->qual &= ~STORAGE_mask;

    if (chop_is_primitive(type)) {
        switch ((int)type->type) {
        case PRIM_int: return rb_str_new2("int");
        case PRIM_void: return rb_str_new2("void");
        case PRIM_char: return rb_str_new2("char");
        case PRIM_short: return rb_str_new2("short");
        case PRIM_long: return rb_str_new2("long");
        case PRIM_float: return rb_str_new2("float");
        case PRIM_double: return rb_str_new2("double");
        case PRIM_schar: return rb_str_new2("signed char");
        case PRIM_uchar: return rb_str_new2("unsigned char");
        case PRIM_ushort: return rb_str_new2("unsigned short");
        case PRIM_uint: return rb_str_new2("unsigned int");
        case PRIM_ulong: return rb_str_new2("unsigned long");
        case PRIM_longlong: return rb_str_new2("long long");
        case PRIM_ulonglong: return rb_str_new2("unsigned long long");
        case PRIM_longdouble: return rb_str_new2("long double");
        default:
            /* should raise error */
            return Qnil;
        }
    }
    else if (type->qual >> 8 == TYPE_typedef >> 8) {
        return type2str(type->type);
    }
    else if (type->qual >> 8 == TYPE_pointer >> 8) {
        VALUE r = type2str(type->type);
        if (r == Qnil) return rb_str_new2("void*");
        rb_str_cat2(r, "*");
        return r;
    }
    else {
        return Qnil;
    }
}

static VALUE rb_get_struct(VALUE k, VALUE s) {
    Sym* sym;
    StringValue(s);

    sym = chop_get_symbol(RSTRING(s)->ptr);
    if (!sym) return Qnil;
    if (!sym->str) return Qnil;

    VALUE mems = rb_ary_new();
    Member* m = sym->str->mems;
    for (; m; m = m->next) {
        /* ignoring enum! */
        VALUE mem = type2str(m->type);
        rb_str_cat2(mem, " ");
        rb_str_cat2(mem, m->name);
        rb_ary_push(mems, mem);
    }
    return mems;
}

static VALUE rb_get_symbol(VALUE k, VALUE s) {
    Sym* sym;
    Type* type;
    StringValue(s);

    sym = chop_get_symbol(RSTRING(s)->ptr);
    if (!sym) return Qnil;
    type = sym->sym;
    if (!type) return Qnil;

    if (type->qual >> 8 == TYPE_function >> 8) {
        Func* func = (Func*)type->type;
        Arg* a = func->args;

        VALUE retargs = rb_ary_new();
        rb_ary_push(retargs, type2str(func->ret));

        for (; a; a = a->next) {
            rb_ary_push(retargs, type2str(a->type));
        }

        return retargs;
    }
    else {
        // not implemented
        return Qnil;
    }
}

void Init_chop() {
    VALUE module;
    module = rb_define_module("Chop");
    rb_define_module_function(module, "c_include", rb_include, 1);
    rb_define_module_function(module, "get_symbol", rb_get_symbol, 1);
    rb_define_module_function(module, "get_struct", rb_get_struct, 1);

    chop_init();
}
